@@ -1,19 +1,8 @@
 /*
- *  Copyright 2020 Marco Martin <mart@kde.org>
+ *  SPDX-FileCopyrightText: 2020 Marco Martin <mart@kde.org>
+ *  SPDX-FileCopyrightText: 2024 ivan tkachenko <me@ratijas.tk>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  2.010-1301, USA.
+ *  SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include "imagecolors.h"
@@ -21,7 +10,6 @@
 #include <QDebug>
 #include <QFutureWatcher>
 #include <QGuiApplication>
-#include <QTimer>
 #include <QtConcurrentRun>
 
 #include "loggingcategory.h"
@@ -47,15 +35,42 @@
             : static_cast<Kirigami::Platform::PlatformTheme *>(qmlAttachedPropertiesObject<Kirigami::Platform::PlatformTheme>(this, true))->finally();         \
     }
 
+PaletteSwatch::PaletteSwatch()
+{
+}
+
+PaletteSwatch::PaletteSwatch(qreal ratio, const QColor &color, const QColor &contrastColor)
+    : m_ratio(ratio)
+    , m_color(color)
+    , m_contrastColor(contrastColor)
+{
+}
+
+qreal PaletteSwatch::ratio() const
+{
+    return m_ratio;
+}
+
+const QColor &PaletteSwatch::color() const
+{
+    return m_color;
+}
+
+const QColor &PaletteSwatch::contrastColor() const
+{
+    return m_contrastColor;
+}
+
+bool PaletteSwatch::operator==(const PaletteSwatch &other) const
+{
+    return m_ratio == other.m_ratio //
+        && m_color == other.m_color //
+        && m_contrastColor == other.m_contrastColor;
+}
+
 ImageColors::ImageColors(QObject *parent)
     : QObject(parent)
 {
-    m_imageSyncTimer = new QTimer(this);
-    m_imageSyncTimer->setSingleShot(true);
-    m_imageSyncTimer->setInterval(100);
-    /* connect(m_imageSyncTimer, &QTimer::timeout, this, [this]() {
-        generatePalette();
-     });*/
 }
 
 ImageColors::~ImageColors()
@@ -161,6 +176,7 @@ void ImageColors::setSourceItem(QQuickItem *source)
             if (m_window) {
                 connect(m_window, &QWindow::visibleChanged, this, &ImageColors::update);
             }
+            update();
         };
 
         connect(m_sourceItem, &QQuickItem::windowChanged, this, syncWindow);
@@ -176,13 +192,16 @@ QQuickItem *ImageColors::sourceItem() const
 void ImageColors::update()
 {
     if (m_futureImageData) {
+        m_futureImageData->disconnect(this, nullptr);
         m_futureImageData->cancel();
         m_futureImageData->deleteLater();
         m_futureImageData = nullptr;
     }
+
     auto runUpdate = [this]() {
-        QFuture<ImageData> future = QtConcurrent::run([this]() {
-            return generatePalette(m_sourceImage);
+        auto sourceImage{m_sourceImage};
+        QFuture<ImageData> future = QtConcurrent::run([sourceImage = std::move(sourceImage)]() {
+            return generatePalette(sourceImage);
         });
         m_futureImageData = new QFutureWatcher<ImageData>(this);
         connect(m_futureImageData, &QFutureWatcher<ImageData>::finished, this, [this]() {
@@ -199,7 +218,7 @@ void ImageColors::update()
         m_futureImageData->setFuture(future);
     };
 
-    if (!m_sourceItem) {
+    if (!m_sourceItem || !m_sourceItem->window() || !m_sourceItem->window()->isVisible()) {
         if (!m_sourceImage.isNull()) {
             runUpdate();
         } else {
@@ -219,13 +238,14 @@ void ImageColors::update()
     if (m_grabResult) {
         connect(m_grabResult.data(), &QQuickItemGrabResult::ready, this, [this, runUpdate]() {
             m_sourceImage = m_grabResult->image();
-            m_grabResult.clear();
             runUpdate();
         });
+        // Workaround for QTBUG-144544
+        connect(m_grabResult.data(), &QQuickItemGrabResult::ready, m_grabResult.data(), [grabResult = m_grabResult] {}, Qt::SingleShotConnection);
     }
 }
 
-inline int squareDistance(QRgb color1, QRgb color2)
+static inline int squareDistance(QRgb color1, QRgb color2)
 {
     // https://en.wikipedia.org/wiki/Color_difference
     // Using RGB distance for performance, as CIEDE2000 is too complicated
@@ -274,7 +294,6 @@ void ImageColors::positionColorMP(const decltype(ImageData::m_samples) &samples,
     std::vector<decltype(ImageData::m_clusters)> tempClusters(numCore, decltype(ImageData::m_clusters){});
 #pragma omp parallel for
     for (int i = 0; i < numCore; ++i) {
-        decltype(ImageData::m_samples) samplePart;
         const auto beginIt = std::next(samples.begin(), numSamplesPerThread * i);
         const auto endIt = i < numCore - 1 ? std::next(samples.begin(), numSamplesPerThread * (i + 1)) : samples.end();
 
@@ -313,7 +332,7 @@ void ImageColors::positionColorMP(const decltype(ImageData::m_samples) &samples,
 #endif
 }
 
-ImageData ImageColors::generatePalette(const QImage &sourceImage) const
+ImageData ImageColors::generatePalette(const QImage &sourceImage)
 {
     ImageData imageData;
 
@@ -382,14 +401,14 @@ ImageData ImageColors::generatePalette(const QImage &sourceImage) const
             g = g / c;
             b = b / c;
             stat.centroid = qRgb(r, g, b);
-            stat.ratio = qreal(stat.colors.count()) / qreal(imageData.m_samples.count());
+            stat.ratio = std::clamp(qreal(stat.colors.count()) / qreal(imageData.m_samples.count()), 0.0, 1.0);
             stat.colors = QList<QRgb>({stat.centroid});
         } // END omp parallel for
 
         positionColorMP(imageData.m_samples, imageData.m_clusters, numCore);
     }
 
-    std::sort(imageData.m_clusters.begin(), imageData.m_clusters.end(), [this](const ImageData::colorStat &a, const ImageData::colorStat &b) {
+    std::sort(imageData.m_clusters.begin(), imageData.m_clusters.end(), [](const ImageData::colorStat &a, const ImageData::colorStat &b) {
         return getClusterScore(a) > getClusterScore(b);
     });
 
@@ -428,10 +447,7 @@ ImageData ImageColors::generatePalette(const QImage &sourceImage) const
 #pragma omp parallel for ordered
     for (int i = 0; i < imageData.m_clusters.size(); ++i) {
         const auto &stat = imageData.m_clusters[i];
-        QVariantMap entry;
         const QColor color(stat.centroid);
-        entry[QStringLiteral("color")] = color;
-        entry[QStringLiteral("ratio")] = stat.ratio;
 
         QColor contrast = QColor(255 - color.red(), 255 - color.green(), 255 - color.blue());
         contrast.setHsl(contrast.hslHue(), //
@@ -464,7 +480,6 @@ ImageData ImageColors::generatePalette(const QImage &sourceImage) const
                             contrast.lightness() > 128 ? qMin(contrast.lightness() + 20, 255) : qMax(0, contrast.lightness() - 20));
         }
 
-        entry[QStringLiteral("contrastColor")] = contrast;
 #pragma omp ordered
         { // BEGIN omp ordered
             if (first) {
@@ -483,14 +498,14 @@ ImageData ImageColors::generatePalette(const QImage &sourceImage) const
             if (qGray(color.rgb()) < qGray(imageData.m_closestToBlack.rgb())) {
                 imageData.m_closestToBlack = color;
             }
-            imageData.m_palette << entry;
+            imageData.m_palette << PaletteSwatch(stat.ratio, color, contrast);
         } // END omp ordered
     }
 
     return imageData;
 }
 
-double ImageColors::getClusterScore(const ImageData::colorStat &stat) const
+double ImageColors::getClusterScore(const ImageData::colorStat &stat)
 {
     return stat.ratio * ColorUtils::chroma(QColor(stat.centroid));
 }
@@ -552,7 +567,7 @@ void ImageColors::postProcess(ImageData &imageData) const
     adjustLightness(imageData.m_average);
 }
 
-QVariantList ImageColors::palette() const
+QList<PaletteSwatch> ImageColors::palette() const
 {
     if (m_futureImageData) {
         qCWarning(KirigamiLog) << m_futureImageData->future().isFinished();
